from pydantic import BaseModel
from decimal import *
from fxpmath import Fxp
from typing import Type

FloatingPoint = Decimal
FixedPoint = Type[Fxp(None, signed=True, n_int=256, n_frac=18)]

class FloatingPointUnstakeTKN(BaseModel):
    a: FloatingPoint
    b: FloatingPoint
    c: FloatingPoint
    e: FloatingPoint
    m: FloatingPoint
    n: FloatingPoint
    x: FloatingPoint
    w: FloatingPoint
    staked_bnt: FloatingPoint
    bnbnt_supply: FloatingPoint

class FixedPointUnstakeTKN(BaseModel):
    a: FixedPoint
    b: FixedPoint
    c: FixedPoint
    e: FixedPoint
    m: FixedPoint
    n: FixedPoint
    x: FixedPoint
    w: FixedPoint
    staked_bnt: FixedPoint
    bnbnt_supply: FixedPoint

FixedPoint = Type[Fxp(None, signed=True, n_int=256, n_frac=18)]

class FixedPointGeneral(BaseModel):
    val: FixedPoint

class FloatingPointGeneral(BaseModel):
    val: FloatingPoint




"""Component that emulates a Staking Ledger."""
from .ledger_base import LedgerBase

class StakingLedger(LedgerBase):
    def __init__(self):
        super().__init__()
        self._ledger = {'block_num': [0]}



from abc import ABC, abstractmethod
from decimal import *


class TokenomicsBase(ABC):
    """Tokenomics base class.

    TODO: Fill-in missing comments below
    """

    @property
    def surplus(self) -> bool:
        """
        """
        return self._surplus

    @surplus.setter
    def surplus(self, is_surplus: bool):
        """
        """
        self._surplus = is_surplus

    @property
    def deficit(self) -> bool:
        """
        """
        return self._deficit

    @deficit.setter
    def deficit(self, is_deficit: bool):
        """
        """
        self._deficit = is_deficit

    @property
    def hlim(self) -> int or Decimal:
        """
        """
        return self._hlim

    @hlim.setter
    def hlim(self, val: int or Decimal):
        """
        """
        self._hlim = val

    @property
    def satisfies_hlim(self) -> bool:
        """
        """
        return self._satisfies_hlim

    @satisfies_hlim.setter
    def satisfies_hlim(self, val: bool):
        """
        """
        self._satisfies_hlim = val

    @property
    def hmax(self) -> int or Decimal:
        """
        """
        return self._hmax

    @hmax.setter
    def hmax(self, val: int or Decimal):
        """
        """
        self._hmax = val

    @property
    def satisfies_hmax(self) -> bool:
        """
        """
        return self._satisfies_hmax

    @satisfies_hmax.setter
    def satisfies_hmax(self, val: bool):
        """
        """
        self._satisfies_hmax = val

    @property
    def reduce_trading_liquidity(self) -> bool:
        """
        """
        return self._reduce_trading_liquidity

    @reduce_trading_liquidity.setter
    def reduce_trading_liquidity(self, val: bool):
        """
        """
        self._reduce_trading_liquidity = val

    @abstractmethod
    def check_surplus(self, b, c, e, n, is_surplus=False, is_deficit=True, case='check surplus'):
        """
        """
        pass

    @abstractmethod
    def check_hmax(self, b, c, e, m, n, x, is_surplus):
        """
        """
        pass

    @abstractmethod
    def check_reduce_trading_liquidity(self, x, b, c, e, n, is_surplus, satisfies_hmax, satisfies_hlim, case='check reduce_trading_liquidity', reduce_trading_liquidity=False):
        """
        """
        pass

    @abstractmethod
    def handle_bootstrap_surplus(self, x, n):
        """
        """
        pass

    @abstractmethod
    def handle_bootstrap_deficit_special_case(self, c, e, n, x):
        """
        """
        pass

    @abstractmethod
    def handle_arbitrage_surplus(self, x, n, a, b, c, e, m):
        """
        """
        pass

    @abstractmethod
    def handle_default_surplus(self, x, n, c, a, b):
        """
        """
        pass

    @abstractmethod
    def handle_arbitrage_deficit(self, x, n, a, m, e, b, c):
        """
        """
        pass

    @abstractmethod
    def handle_bootstrap_deficit(self, x, n, b, c, e, a):
        """
        """
        pass

    @abstractmethod
    def handle_default_deficit(self, a, b, c, e, n, x):
        """
        """
        pass

    @abstractmethod
    def handle_external_wallet_adjustment_1(self, t, b, a):
        """
        """
        pass

    @abstractmethod
    def handle_external_wallet_adjustment_2(self, w, t, b, a):
        """
        """
        pass

    def check_external_wallet_adjustment(self, a, b, w, t, satisfies_hlim, satisfies_hmax, case=None):
        if ((satisfies_hlim == False) or (satisfies_hmax == False)) & (w > 0):
            if (a * w) > (t * b):
                case = 'external wallet adjustment (1)'
            else:
                case = 'external wallet adjustment (2)'
        return case








"""Component that emulates a Available Liquidity Ledger."""
from .ledger_base import LedgerBase


class AvailableLiquidityLedger(LedgerBase):
    """Component that emulates a Available Liquidity Ledger."""
    def __init__(self):
        super().__init__()
        self._ledger = {
            'LINK': {
                'block_num': [0],
                'funding_limit': [0],
                'BNT': [0],
                'LINK': [0]
            },
            'ETH': {
                'block_num': [0],
                'funding_limit': [0],
                'BNT': [0],
                'ETH': [0]
            },
            'DAI': {
                'block_num': [0],
                'funding_limit': [0],
                'BNT': [0],
                'DAI': [0]
            },
            'wBTC': {
                'block_num': [0],
                'funding_limit': [0],
                'BNT': [0],
                'wBTC': [0]
            },
            'TKN': {
                'block_num': [0],
                'funding_limit': [0],
                'BNT': [0],
                'TKN': [0]
            },
            'BNT': {
                'block_num': [0],
                'funding_limit': [0],
                'BNT': [0]
            },
        }




"""Component that emulates a Liquidity Pool Ledger."""
from .ledger_base import LedgerBase

class PoolTokenSupplyLedger(LedgerBase):
    """Component that emulates a Liquidity Pool Ledger."""
    def __init__(self):
        super().__init__()
        self._ledger = {}




"""Load the datasets"""
import pandas as pd
from decimal import *
import os

def decimal_from_value(value):
    return Decimal(value)

def parse_json_tests(X, indx):
    a = Decimal(X['a'].values[indx])
    b = Decimal(X['b'].values[indx])
    c = Decimal(X['c'].values[indx])
    e = Decimal(X['e'].values[indx])
    n = Decimal(X['n'].values[indx])
    x = Decimal(X['x'].values[indx])
    w = Decimal(X['w'].values[indx])
    m = Decimal(X['m'].values[indx])
    p = Decimal(X['P'].values[indx])
    q = Decimal(X['Q'].values[indx])
    r = Decimal(X['R'].values[indx])
    s = Decimal(X['S'].values[indx])
    t = Decimal(X['T'].values[indx])
    u = Decimal(X['U'].values[indx])
    return a, b, c, e, n, x, w, m, p, q, r, s, t, u

converters={'a': str,
           'b': str,
           'c': str,
           'e': str,
           'm': str,
           'n': str,
           'w': str,
           'x': str,
           'is_tkn_surplus': str,
           'hlim': str,
           'satisfies_hlim': str,
           'hmax': str,
           'satisfies_hmax': str,
           'require_reduce_trading_liquidity': str,
           'P': str,
           'Q': str,
           'P': str,
           'Q': str,
           'R': str,
           'S': str,
           'T': str,
           'U': str,
}

def load_test_data():
    data_path = os.path.join(os.path.dirname(__file__), "data/withdrawal_algorithm_tests.json")
    data_path = data_path.replace('utils', 'tests/environment_tests')
    return pd.read_json(data_path, dtype=converters)

def load_data(path='data/MBRtest1.json', filetype='json', index=None, target=None, n_rows=None, drop=None, verbose=True, **kwargs):
    """Load features and target from file.

    Args:
        path (str): Path to file or a http/ftp/s3 URL.
        index (str): Column for index.
        target (str): Column for target.
        n_rows (int): Number of rows to return. Defaults to None.
        drop (list): List of columns to drop. Defaults to None.
        verbose (bool): If True, prints information about features and target. Defaults to True.
        **kwargs: Other keyword arguments that should be passed to panda's `read_csv` method.

    Returns:
        pd.DataFrame, pd.Series: Features matrix and target.
    """
    if filetype=='csv':
        feature_matrix = pd.read_csv(path, index_col=index, nrows=n_rows, dtype=converters, **kwargs)
    elif filetype=='json':
        feature_matrix = pd.read_json(path, index_col=index, nrows=n_rows, dtype=converters, **kwargs)

    if target is not None:
        targets = [target] + (drop or [])
        y = feature_matrix[target]
        X = feature_matrix.drop(columns=targets)
    else:
        X = feature_matrix

    if verbose:
        # number of features
        print(len(X.columns), end="\n\n")

        # number of total training examples
        info = "Number of training examples: {}"
        print(info.format(len(X)), end="\n")

    try: return X, y
    except: return X



import unittest
import pytest
from bancorml.utils import parse_json_tests, load_test_data
from bancorml.environments import Bancor3
from bancorml.utils.schemas import FloatingPointUnstakeTKN

PRECISION = 10
IS_SOLIDITY = False
protocol = Bancor3(is_solidity=IS_SOLIDITY)

@pytest.mark.usefixtures("protocol")
class TestWithdrawalAlgorithm(unittest.TestCase):

    def test_check_surplus(self):
        X = load_test_data()
        for indx in range(len(X)):
            a, b, c, e, n, x, w, m, _p, _q, _r, _s, _t, _u = parse_json_tests(X, indx)
            t_is_surplus = True if X['is_tkn_surplus'].values[indx]=='True' else False
            protocol = Bancor3(is_solidity=IS_SOLIDITY)
            protocol.tokenomics.check_surplus(b, c, e, n)
            is_surplus, is_deficit = protocol.tokenomics.surplus, protocol.tokenomics.deficit
            assert is_surplus == t_is_surplus, f"is_surplus={is_surplus} != t_is_surplus={t_is_surplus} on index={indx}"

    def test_hlim(self):
        X = load_test_data()
        for indx in range(len(X)):
            a, b, c, e, n, x, w, m, _p, _q, _r, _s, _t, _u = parse_json_tests(X, indx)
            t_satisfies_hlim = True if X['satisfies_hlim'].values[indx]=='True' else False
            protocol.tokenomics.check_hlim(b, c, e, x)
            satisfies_hlim = protocol.tokenomics.satisfies_hlim
            assert satisfies_hlim == t_satisfies_hlim, f"test_hlim failed on index={indx}"

    def test_hmax_deficit(self):
        X = load_test_data()
        X = X[(X['is_tkn_surplus'] == 'False') & (X['x'].astype(float) > 0)]
        for indx in range(len(X)):
            a, b, c, e, n, x, w, m, _p, _q, _r, _s, _t, _u = parse_json_tests(X, indx)
            t_satisfies_hmax = True if X['satisfies_hmax'].values[indx]=='True' else False
            protocol.tokenomics.check_hmax(b, c, e, m, n, x, False)
            satisfies_hmax  = protocol.tokenomics.satisfies_hmax
            assert satisfies_hmax == t_satisfies_hmax, f"test_hmax_deficit failed on index={indx}"

    def test_hmax_surplus(self):
        X = load_test_data()
        X = X[(X['is_tkn_surplus'] == 'True') & (X['x'].astype(float) > 0)]
        for indx in range(len(X)):
            a, b, c, e, n, x, w, m, _p, _q, _r, _s, _t, _u = parse_json_tests(X, indx)
            t_satisfies_hmax = True if X['satisfies_hmax'].values[indx] == 'True' else False
            protocol.tokenomics.check_hmax(b, c, e, m, n, x, True)
            satisfies_hmax = protocol.tokenomics.satisfies_hmax
            assert satisfies_hmax == t_satisfies_hmax, f"test_hmax_surplus failed on index={indx}"

    def test_reduce_trading_liquidity(self):
        X = load_test_data()
        X = X[X['x'].astype(float) > 0]
        for indx in range(len(X)):
            a, b, c, e, n, x, w, m, _p, _q, _r, _s, _t, _u = parse_json_tests(X, indx)
            satisfies_hmax = True if X['satisfies_hmax'].values[indx] == 'True' else False
            is_surplus = True if X['is_tkn_surplus'].values[indx] == 'True' else False
            satisfies_hlim = True if X['satisfies_hlim'].values[indx] == 'True' else False
            require_reduce_trading_liquidity = True if X['require_reduce_trading_liquidity'].values[indx] == 'True' else False
            protocol.tokenomics.check_reduce_trading_liquidity(x, b, c, e, n, is_surplus, satisfies_hmax, satisfies_hlim)
            reduce_trading_liquidity = protocol.tokenomics.reduce_trading_liquidity
            assert reduce_trading_liquidity == require_reduce_trading_liquidity, f"test_reduce_trading_liquidity failed on index={indx}"

    def test_p(self):
        X = load_test_data()
        X = X[X['x'].astype(float) > 0]
        for indx in range(len(X)):
            a, b, c, e, n, x, w, m, _p, _q, _r, _s, _t, _u = parse_json_tests(X, indx)
            input_params = dict(
                a=a,
                b=b,
                e=e,
                w=w,
                c=c,
                n=n,
                m=m,
                x=x
            )
            params = FloatingPointUnstakeTKN(**input_params)
            case, p, q, r, s, t, u = protocol.get_state(params)
            if 'E' not in str(p) and 'E' not in str(_p):
                p = str(p).split('.')[0] + '.' + str(p).split('.')[-1][:PRECISION]
                _p = str(_p).split('.')[0] + '.' + str(_p).split('.')[-1][:PRECISION]
                assert p == _p, f"test_p failed on index={indx}, case={case}"

    def test_q(self):
        X = load_test_data()
        X = X[X['x'].astype(float) > 0]
        for indx in range(len(X)):
            a, b, c, e, n, x, w, m, _p, _q, _r, _s, _t, _u = parse_json_tests(X, indx)
            input_params = dict(
                a=a,
                b=b,
                e=e,
                w=w,
                c=c,
                n=n,
                m=m,
                x=x
            )
            params = FloatingPointUnstakeTKN(**input_params)
            case, p, q, r, s, t, u = protocol.get_state(params)
            if 'E' not in str(q) and 'E' not in str(_q):
                q = str(q).split('.')[0] + '.' + str(q).split('.')[-1][:PRECISION]
                _q = str(_p).split('.')[0] + '.' + str(_q).split('.')[-1][:PRECISION]
                assert q == _q, f"test_q failed on index={indx}, case={case}"

    def test_r(self):
        X = load_test_data()
        X = X[X['x'].astype(float) > 0]
        for indx in range(len(X)):
            a, b, c, e, n, x, w, m, _p, _q, _r, _s, _t, _u = parse_json_tests(X, indx)
            input_params = dict(
                a=a,
                b=b,
                e=e,
                w=w,
                c=c,
                n=n,
                m=m,
                x=x
            )
            params = FloatingPointUnstakeTKN(**input_params)
            case, p, q, r, s, t, u = protocol.get_state(params)
            if 'E' not in str(r) and 'E' not in str(_r):
                r = str(r).split('.')[0] + '.' + str(r).split('.')[-1][:PRECISION]
                _r = str(_r).split('.')[0] + '.' + str(_r).split('.')[-1][:PRECISION]
                assert r == _r, f"test_q failed on index={indx}, case={case}"

    def test_s(self):
        X = load_test_data()
        X = X[X['x'].astype(float) > 0]
        for indx in range(len(X)):
            a, b, c, e, n, x, w, m, _p, _q, _r, _s, _t, _u = parse_json_tests(X, indx)
            input_params = dict(
                a=a,
                b=b,
                e=e,
                w=w,
                c=c,
                n=n,
                m=m,
                x=x
            )
            params = FloatingPointUnstakeTKN(**input_params)
            case, p, q, r, s, t, u = protocol.get_state(params)
            if 'E' not in str(s) and 'E' not in str(_s):
                s = str(s).split('.')[0] + '.' + str(s).split('.')[-1][:PRECISION]
                _s = str(_s).split('.')[0] + '.' + str(_s).split('.')[-1][:PRECISION]
                assert s == _s, f"test_q failed on index={indx}, case={case}"

    def test_t(self):
        X = load_test_data()
        X = X[X['x'].astype(float) > 0]
        for indx in range(len(X)):
            a, b, c, e, n, x, w, m, _p, _q, _r, _s, _t, _u = parse_json_tests(X, indx)
            input_params = dict(
                a=a,
                b=b,
                e=e,
                w=w,
                c=c,
                n=n,
                m=m,
                x=x
            )
            params = FloatingPointUnstakeTKN(**input_params)
            case, p, q, r, s, t, u = protocol.get_state(params)
            if 'E' not in str(t) and 'E' not in str(_t):
                t = str(t).split('.')[0] + '.' + str(t).split('.')[-1][:PRECISION]
                _t = str(_t).split('.')[0] + '.' + str(_t).split('.')[-1][:PRECISION]
                assert t == _t, f"test_t failed on index={indx}, case={case}"

    def test_u(self):
        X = load_test_data()
        X = X[X['x'].astype(float) > 0]
        for indx in range(len(X)):
            a, b, c, e, n, x, w, m, _p, _q, _r, _s, _t, _u = parse_json_tests(X, indx)
            input_params = dict(
                a=a,
                b=b,
                e=e,
                w=w,
                c=c,
                n=n,
                m=m,
                x=x
            )
            params = FloatingPointUnstakeTKN(**input_params)
            case, p, q, r, s, t, u = protocol.get_state(params)
            if 'E' not in str(u) and 'E' not in str(_u):
                u = str(u).split('.')[0] + '.' + str(u).split('.')[-1][:PRECISION]
                _u = str(_u).split('.')[0] + '.' + str(_u).split('.')[-1][:PRECISION]
                assert u == _u, f"test_u failed on index={indx}, case={case}"



from abc import ABC, abstractmethod

class AgentBase(ABC):
    """Base class for all agents."""

    @property
    @classmethod
    @abstractmethod
    def name(cls):
        """Returns string name of this component."""




from bancorml.environments.tokenomics.tokenomics_base import TokenomicsBase
from decimal import *
import numpy as np

class FloatingPointTokenomics(TokenomicsBase):
    """Tokenomics subclass for all floating-point arithmetic.
    """

    def handle_unstake_bnt(self, x, exchange_rate, exit_fee, tkn_out=0):
        bnt = exchange_rate * x
        exit_fee = exit_fee * bnt
        bnt_out = bnt - exit_fee
        return tkn_out, bnt_out

    def handle_division(self, numerator, denominator):
        exchange_rate = numerator / denominator
        return exchange_rate

    def handle_multiplication(self, x1, x2):
        return x1 * x2

    def handle_addition(self, x1, x2):
        return x1 + x2

    def handle_subtraction(self, x1, x2):
        return x1 - x2

    def handle_ema(self, alpha, spot_rate, current_ema):
        return alpha * spot_rate + current_ema * (1 - alpha)

    def get_pool_tokens_issued_to_user(self, x, bnbnt_supply, staked_bnt):
        exchange_rate = bnbnt_supply / staked_bnt
        pool_tokens_issued_to_user = exchange_rate * x
        return pool_tokens_issued_to_user

    def check_balanced(self, b, c, e, is_surplus=False, is_deficit=True, is_balanced=False):
        if b + c - e == 0:
            is_balanced, is_deficit, is_surplus = True, False, False
        self.is_balanced, self.surplus, self.deficit, self.hlim, self.satisfies_hlim, self.hmax, self.satisfies_hmax, self.reduce_trading_liquidity = is_balanced, is_surplus, is_deficit, np.nan, np.nan, np.nan, np.nan, np.nan

    def check_surplus(self, b, c, e, n, is_surplus=False, is_deficit=True, case='check surplus'):
        if (b + c) / (1 - n) > e: is_surplus, is_deficit = True, False
        self.surplus, self.deficit = is_surplus, is_deficit

    def check_hlim(self, b, c, e, x):
        hlim = (c * e) / (b + c)
        satisfies_hlim = hlim > x
        self.hlim, self.satisfies_hlim = hlim, satisfies_hlim

    def check_hmax(self, b, c, e, m, n, x, is_surplus):
        if is_surplus:
            hmax = b * e * (e * n + m * (b + c - e)) / ((1 - m) * (b + c - e) * (b + c - e * (1 - n)))
        else:
            hmax = b * e * (e * n + m * (e * (1 - n) - b - c)) / ((1 - m) * (e - b - c) * (e * (1 - n) - b - c))

        satisfies_hmax = hmax > x
        self.hmax, self.satisfies_hmax = hmax, satisfies_hmax

    def check_reduce_trading_liquidity(self, x, b, c, e, n, is_surplus, satisfies_hmax, satisfies_hlim,
                                       case='check reduce_trading_liquidity', reduce_trading_liquidity=False):
        if (satisfies_hmax == False) or (satisfies_hlim == False):
            reduce_trading_liquidity = (x * (1 - n)) <= c if is_surplus else ((x * (1 - n) * (b + c)) / e) <= c
        self.reduce_trading_liquidity = reduce_trading_liquidity

    def handle_bootstrap_surplus(self, x, n):
        s = x * (1 - n)
        return s

    def handle_bootstrap_deficit_special_case(self, c, e, n, x):
        s = c * x * (1 - n) / e
        return s

    def handle_arbitrage_surplus(self, x, n, a, b, c, e, m):
        s = x * (1 - n)
        p = a * x * (b + c - e * (1 - n)) / ((1 - m) * (b * e + x * (b + c - e * (1 - n))))
        r = x * (b + c - e * (1 - n)) / e
        return p, r, s

    def handle_default_surplus(self, x, n, c, a, b):
        r = x * (1 - n) - c
        s = x * (1 - n)
        p = q = a * (x * (1 - n) - c) / b
        return p, q, r, s

    def handle_arbitrage_deficit(self, x, n, a, m, e, b, c):
        s = x * (1 - n)
        p = a * x * (1 - m) * (e * (1 - n) - b - c) / (b * e - x * (1 - m) * (e * (1 - n) - b - c))
        r = x * (e * (1 - n) - b - c) / e
        return p, r, s

    def handle_bootstrap_deficit(self, x, n, b, c, e, a):
        s = x * (1 - n) * (b + c) / e
        t = a * x * (1 - n) * (e - b - c) / (b * e)
        return s, t

    def handle_default_deficit(self, a, b, c, e, n, x):
        p = q = (a * (b * e - (b + c) * (e - x * (1 - n)))) / (b * e)
        r = (x * (1 - n) * (b + c) - c * e) / e
        s = (x * (1 - n) * (b + c)) / e
        t = (a * x * (1 - n) * (e - b - c)) / (b * e)
        return p, q, r, s, t

    def handle_external_wallet_adjustment_1(self, t, b, a):
        u = (t * b) / a
        t = 0
        return t, u

    def handle_external_wallet_adjustment_2(self, w, t, b, a):
        u = w
        t = (t * b - a * w) / b
        return t, u

    def handle_trade_bnt_to_tkn(self, a, d, b, x, e):
        bnt_trading_liquidity = (a * (a + x) + d * (1 - e) * (a * x + x * x)) / (a + d * x)
        tkn_trading_liquidity = (b * (a + d * x)) / (a + x)
        tkn_out = (b * x * (1 - d)) / (a + x)
        tkn_fee = (b * d * x * (1 - e)) / (a + x)
        vortex_fee = (d * e * x * (a + x)) / (a + d * x)
        return bnt_trading_liquidity, tkn_trading_liquidity, tkn_out, tkn_fee, vortex_fee

    def handle_trade_tkn_to_bnt(self, a, d, b, x, e):
        print(f'a={a}, d={d}, b={b}, x={x}, e={e}')
        bnt_trading_liquidity = (a * (b + d * x * (1 - e))) / (b + x)
        tkn_trading_liquidity = b + x
        bnt_out = (a * x * (1 - d)) / (b + x)
        bnt_fee = (a * d * x * (1 - e)) / (b + x)
        vortex_fee = (a * d * e * x) / (b + x)
        return bnt_trading_liquidity, tkn_trading_liquidity, bnt_out, bnt_fee, vortex_fee

    def handle_trade_tkn_to_tkn(self, a1, d1, b1, a2, d2, b2, x, e):
        bnt_source_trading_liquidity = (a1 * (b1 + d1 * x * (1 - e))) / (b1 + x)
        tkn_source_trading_liquidity = b1 + x
        bnt_destination_trading_liquidity = ((a1 * x * (1 - d1) + a2 * (b1 + x)) * (a1 * d2 * x * (1 - d1) * (1 - e) + a2 * (b1 + x))) / ((b1 + x) * (a1 * d2 * x * (1 - d1) + a2 * (b1 + x)))
        tkn_destination_trading_liquidity = (b2 * (a1 * d2 * x * (1 - d1) + a2 * (b1 + x))) / (a1 * x * (1 - d1) + a2 * (b1 + x))
        tkn_out = (a1 * b2 * x * (1 - d1) * (1 - d2)) / (a1 * x * (1 - d1) + a2 * (b1 + x))
        bnt_fee = (a1 * d1 * x * (1 - e)) / (b1 + x)
        tkn_fee = (a1 * b2 * d2 * x * (1 - d1) * (1 - e)) / (a1 * x * (1 - d1) + a2 * (b1 + x))
        vortex_fee = (a1 * e * x * (a1 * d2 * x * (1 - d1) + a2 * (b1 + x) * (d1 + d2 - d1 * d2))) / ((b1 + x) * (a1 * d2 * x * (1 - d1) + a2 * (b1 + x)))
        return bnt_source_trading_liquidity, tkn_source_trading_liquidity, bnt_destination_trading_liquidity, tkn_destination_trading_liquidity, tkn_out, bnt_fee, tkn_fee, vortex_fee
